/* 
 *	Driver pro FDC - WD279[1/3]
 *
 *	Verze: 0.1 (31.12.2011)
 *
 *	Copyright (c) 2011 by Michal Hucik <http://www.ordoz.com>
 *
*/

#include "driver_FDC.h"

#define	DI	__asm di __endasm
#define	EI	__asm ei __endasm
#define	IM1	__asm im 1 __endasm


/* porty WD279x */
__sfr __at	0xd8	FDC_CMD;	/* out */
__sfr __at	0xd8	FDC_STS;	/* in */
__sfr __at	0xd9	FDC_TRACK;	/* in, out */
__sfr __at	0xda	FDC_SECTOR;	/* in, out */
__sfr __at	0xdb	FDC_DATA;	/* in, out */
__sfr __at	0xdc	FDC_DRIVE;	/* out, motor + drive */
__sfr __at	0xdd	FDC_SIDE;	/* out, 0. bit */
__sfr __at	0xde	FDC_DENSITY;	/* out, 0. bit */
__sfr __at	0xdf	FDC_EINT;	/* out, 0. bit */


/* FDC_CMD (jiz v obracene bitove logice!) */



/* prikazy typ I. */
#define CMD_RESTORE	0xf0
#define CMD_SEEK	0xe0
#define CMD_STEP_IN	0xa0
#define CMD_STEP_OUT	0x80

/* overovani stopy u prikazu STEP, SPTEP-IN a STEP-OUT */
#define DEF_TRACK	( 0 << 4 )	/* definovano registrem stopy */
#define NOT_DEF_TRACK	( 1 << 4 )

#define USE_HEAD	( 0 << 3 )	/* hlava prilozena k mediu */
#define NOT_USE_HEAD	( 1 << 3 )

#define VERIFY_TRK	( 0 << 2 )	/* overit ctenim cisla stopy */
#define NO_VERIFY_TRK	( 1 << 2 )

/* Head speed - bity 0 a 1 */
#define HSPEED_6MS	0x03
#define HSPEED_12MS	0x02
#define HSPEED_20MS	0x01
#define HSPEED_30MS	0x00


#define NOT_USE_HEAD	( 1 << 3 )


/* prikazy typ II. */
#define CMD_READ_SEC	0x60
#define CMD_WRITE_SEC	0x40

#define MULTI_SECTOR	( 0 << 4 )	/* cteni/zapis az do konce stopy */
#define SINGLE_SECTOR	( 1 << 4 )

/* lze pouzit u prikazu II. i III. typu */
#define DELAY15MS	( 0 << 2 )
#define NO_DELAY15MS	( 1 << 2 )

/* pokud overujeme stranu, tak je uvedena ve 3. bitu ! */
#define VERIFY_SIDE	( 0 << 1 )
#define NO_VERIFY_SIDE	( 1 << 1 )

/* Znacka adresy dat - zatim netusim k cemu tohle slouzi, ale asi je to jen pro zapis? */
#define DATA_SIGN_FB	( 1 << 0 )	/* 0FBh (DAM) */
#define DATA_SIGN_F8	( 0 << 0 )	/* 0F8h (vypustena DAM) */


/* prikazy typ III. */

#define CMD_READ_ADDR	0x3b
#define CMD_READ_TRACK	0x1b
#define CMD_WRITE_TRACK	0x0b

/* prikazy typ IV. */
#define CMD_BREAK	0x2f
#define CMD_RESET	0x27



/* FDC_DRIVE */
#define MOTOR_ON	( 1 << 7 )
#define MOTOR_OFF	0x00

#define DRIVE_SELECT	( 1 << 2 )


/* FDC_EINT */
#define EINT_ON		0x01
#define EINT_OFF	0x00






/* prave nastavena strana (invertovano a posunuto na 3.bit pro pouziti pri READ a WRITE) */
volatile unsigned char gSIDE = 0;





/* delay34us - konstantni cekaci smycka, ktera neovlivnije registry (cca 34us)
 *
*/
void delay34us ( void ) __naked {

    __asm
    
    ex (sp), hl
    ex (sp), hl
    ex (sp), hl
    ex (sp), hl
    ex (sp), hl
    ex (sp), hl
    ret
    
    __endasm;
}



/*
 * FDC_waitForNotBusy - cekame dokud z FDC_STS nezmizi priznak BUSY
 *
 * Vraci:
 *	-1: nedockali jsme se
 *	 0: NOT BUSY
 *
 * Pozn.: musi zachovat C (u STEP_IN a STEP_OUT tam mame command)!
*/
char FDC_waitForNotBusy ( void ) __naked {

    __asm
    
    ld l,#0x00
    ld b,#0x02
    
    1$:
    call _delay34us
    ld de,#0x0000
    
    2$:
    in a, (_FDC_STS)
    rrc a
    ret c		// neni BUSY, return ( 0 )
    dec de
    ld a,d
    or e
    jr nz, 2$

    dec b
    jr z,3$
    ld a,#CMD_BREAK
    out (_FDC_CMD),a
    jr 1$
    
    3$:
    dec l		// return ( -1 )
    ret
    
    __endasm;

/*
    unsigned char attempt = 2;
    unsigned int timeout;
    
    while ( attempt-- ) {
        delay34us ();
        timeout = 0xffff;
        while ( timeout-- ) {
            if ( FDC_STS & FDC_STS_BUSY ) return ( 0 );	// ( STATUS != BUSY ) => OK
        };
        // nedockali jsme se, tak zkusime zastaveni prikazu bez preruseni
        FDC_CMD = CMD_BREAK;
    };
    return ( -1 );
*/
}





/*
 * FDC_waitForFinishCommand - po odeslani prikazu cekame dokud neni zpracovan
 *			      (zmizi NOT_RDY)
 *
 * Vraci:
 *	0: OK, nebo -1 pri chybe
 *
*/
char FDC_waitForFinishCommand ( void ) __naked {

    __asm
    
    ld l,#0x00
    call _delay34us
    ld de,#0x0000
    
    1$:
    in a, (_FDC_STS)
    and #FDC_STS_NOT_RDY | FDC_STS_BUSY
    xor #FDC_STS_NOT_RDY
    ret z		// neni NOT READY a je BUSY, return ( 0 )
    dec de
    ld a,d
    or e
    jr nz, 1$

    dec l		// return ( -1 )
    ret
    
    __endasm;

/*
    unsigned int timeout;
    
    delay34us ();
    timeout = 0xffff;
    while ( ( FDC_STS & ( FDC_STS_NOT_RDY | FDC_STS_BUSY ) ) ^ FDC_STS_NOT_RDY ) {
        if ( ! timeout-- ) return ( -1 ); // nedockali jsme se
    };
    return ( 0 ); // pochroustani prikazu dokonceno
*/
}




/*
 *
 * Rutiny volane pri DRQ
 *
 * Pozn.: nesmi ovlivnit A, a Carry !
*/


/*
 * _FDC_readByte - DRQ handler pro neinvertovane cteni
 *
 *
*/
void _FDC_readByte ( void ) __naked {
    __asm
    ini
#ifdef FDC_SUPPORT_HDD
    ei
    reti
#else
    ret
#endif
    __endasm;
}


/*
 * _FDC_readByteCpl - DRQ handler pro invertovane cteni
 *
 *
*/
void _FDC_readByteCpl ( void ) __naked {
    __asm
#ifdef FDC_SUPPORT_HDD
    ld d,a
#endif
    in a,(c)
    cpl
    ld (hl),a
    inc hl
#ifdef FDC_SUPPORT_HDD
    ld a,d
    ei
    reti
#else
    ret
#endif
    __endasm;
}




/*
 * _FDC_writeByte - DRQ handler pro neinvertovany zapis
 *
 *
*/
void _FDC_writeByte ( void ) __naked {
    __asm
    outi
#ifdef FDC_SUPPORT_HDD
    ei
    reti
#else
    ret
#endif
    __endasm;
}


/*
 * _FDC_writeByteCpl - DRQ handler pro invertovany zapis
 *
 *
*/
void _FDC_writeByteCpl ( void ) __naked {
    __asm
#ifdef FDC_SUPPORT_HDD
    ld d,a
#endif
    ld a,(hl)
    cpl
    out (c),a
    inc hl
#ifdef FDC_SUPPORT_HDD
    ld a,d
    ei
    reti
#else
    ret
#endif
    __endasm;
}




/*
 * FDC_waitForData - radic je ve stavu, kdy zapisujeme/cteme data
 *		     budeme na ne cekat, dokud bude vystaven priznak BUSY
 *
 * Vstup:
 *	dma_addr - adresa dat
 *	drq_call - rutina, ktera se vykona pri DRQ
 *
 * Vystup:
 *	pocet prectenych bajtu
 *
*/
unsigned int FDC_waitForData ( void* p, void *drq_call ) __naked {

    (void) p;
    (void) drq_call;
    __asm

    push ix
    ld ix,#0
    add ix,sp
    
    // Prepiseme adresu volani pro DRQ.
    //
    // Pri FDC_SUPPORT_HDD zapiseme na 0x0038 instrukci JP <drq_call>
    // U normalniho radice prepiseme CALL intrukci pro volani pri DRQ.
    ld e,6 (ix)
    ld d,7 (ix)

#ifdef FDC_SUPPORT_HDD
    ld hl,#0x0038
    ld a,#0xc3
    ld (hl),a
    inc hl
#else    
    ld hl,#3$+1
#endif
    ld (hl),e
    inc hl
    ld (hl),d
    
    // DMA do HL
    ld l,4 (ix)
    ld h,5 (ix)
    
    ld c,#_FDC_DATA
 
#ifdef FDC_SUPPORT_HDD
    ei
#endif    
    // cekame na DRQ, nebo az zmizi BUSY
    1$:
    in a,(#_FDC_STS)
    rrc a
    jr c, 2$				// zmizel BUSY - koncime

#ifndef FDC_SUPPORT_HDD
    rrc a
    
    3$:
    call nc, __FDC_readByteCpl		// prisel DRQ - jdeme pro dalsi bajt
#endif

    jr 1$
    
    2$:
#ifdef FDC_SUPPORT_HDD
    di
#endif
    // zacatek DMA do DE
    ld e,4 (ix)
    ld d,5 (ix)
    ccf
    sbc hl,de				// do HL pocet prenesenych bajtu
    pop ix
    ret
    
    __endasm;
}





/*
 *
 *   PRIKAZY PRO PRACI S MECHANIKOU
 *
 *
*/



/*
 * FDC_doMotorOn - zapne motor
 *
*/
void FDC_doMotorOn ( void ) __naked {
    FDC_DRIVE = MOTOR_ON;
}





/*
 * FDC_doMotorOff - zastavi motor
 *
*/
void FDC_doMotorOff ( void ) {
    FDC_DRIVE = MOTOR_OFF;
}





/*
 * FDC_doSelectDrive - zapne motor a vybere mechaniku
 *
 * Vstup:
 *	drive = 0 - 3
*/
void FDC_doSelectDrive ( char drive_id ) __naked {
    (void) drive_id;
    
    __asm
    
    pop de
    pop bc
    push bc
    push de

    ld a,#0x03
    and c
    or #MOTOR_ON | DRIVE_SELECT
    out (_FDC_DRIVE),a
    ret

    __endasm;
    
//    FDC_DRIVE = MOTOR_ON | DRIVE_SELECT | ( drive_id & 0x03 );
}





/*
 *
 *   PRACE S REGISTRY FDC
 *
 *
*/




/*
 * FDC_setTrack - nastaveni registru stopy
 *
 *
*/
void FDC_setTrack ( unsigned char track ) __naked {
    (void) track;

    __asm
    
    pop de
    pop bc
    push bc
    push de

    ld a,c
    cpl
    out (_FDC_TRACK),a
    ret
    
    __endasm;
//    FDC_TRACK = ~track;
}




/*
 * FDC_getTrack - cteni registru stopy
 *
 *
*/
unsigned char FDC_getTrack ( void ) {
// hmmm ... tohle chce overit na realnem radici ...
#if 0
// Pozn.: cteni aktualni stopy tak, jak to dela cp/m - nejprve posleme do FDC_TRACK = 0x00 a pak jej precteme.
    __asm
        ld a,#0xff
        out (_FDC_TRACK),a
        // mozna tu bude potreba chvili pockat (?)
        in a,(_FDC_TRACK)
        cpl
        ld l,a
        ret
    __endasm;
#endif
    return ( ~FDC_TRACK );
}




/*
 * FDC_setSector - nastaveni registru sektoru
 *
 *
*/
void FDC_setSector ( unsigned char sector ) __naked {
    (void) sector;

    __asm
    
    pop de
    pop bc
    push bc
    push de

    ld a,c
    cpl
    out (_FDC_SECTOR),a
    ret
    
    __endasm;
//    FDC_SECTOR = ~sector;
}




/*
 * FDC_getSector - cteni registru sector
 *
 *
*/
unsigned char FDC_getSector ( void ) {
    return ( ~FDC_SECTOR );
}





/*
 * FDC_setSide - nastaveni strany (aplkuje se jen 0. bit)
 *
 *
*/
void FDC_setSide ( unsigned char side ) __naked {
    (void) side;

    __asm
    
    pop de
    pop bc
    push bc
    push de

    ld a,c
    out (_FDC_SIDE),a

    cpl
    and #0x01
    rlca
    rlca
    rlca
    ld hl,#_gSIDE
    ld (hl),a
    
    ret
    
    __endasm;

// Tohle se zkompiluje nejak hodne divoce:
//    gSIDE = ( ( ~side ) & 1 ) << 3;	// tuto hodnotu pouzijeme pri overeni strany u RW operaci
//    FDC_SIDE = side;	// neinvertuje se!
}






/*
 * FDC_getStatus - cteni status registru
 *
 *
*/
unsigned char FDC_getStatus ( void ) {
    return ( FDC_STS );	// vraci invertovanou hodnotu!
}





#ifdef FDC_SUPPORT_HDD
/*
 * FDC_setDensity - nastaveni HD/DD (aplkuje se jen 0. bit)
 *		    0: DD
 *		    1: HD (vyzaduje upraveny FDC, ktery pracuje v INT rezimu)
 *
*/
void FDC_setDensity ( unsigned char density ) __naked {
    (void) density;

    __asm
    
    pop de
    pop bc
    push bc
    push de

    ld a,c
    out (_FDC_DENSITY),a
    ret
    
    __endasm;
//    FDC_DENSITY = density; // neinvertuje se!
}
#endif




/*
 *
 *   KROKOVACI PRIKAZY (TYP I.)
 *
 *
*/




/*
 * FDC_doStepInRestore - 
 *
 * Vstup:
 *	reg. C	- command
 *
 * Vraci:
 *	 0: vse OK
 *	-1: prikaz nelze vlozit
 *	-2: prikaz ukoncen s chybou
 */
char _FDC_StepInOutRestore ( void ) __naked {

    __asm
    
    call _FDC_waitForNotBusy
    ret nc	// return ( -1 )
    
    ld a,c
    out (_FDC_CMD), a
    
    call _FDC_waitForNotBusy
    ld a,l
    or a
    ret z	// return ( 0 )

    dec l	// return ( -2 )
    ret
        
    __endasm;
}




/*
 * FDC_doRestore - nastavit stopu 0
 *
 * Vraci:
 *	0: OK, nebo zaporne cislo pri chybe
 *
*/
char FDC_doRestore ( void ) __naked {

    __asm
    
    ld c,#CMD_RESTORE | NO_VERIFY_TRK | HSPEED_30MS
    call __FDC_StepInOutRestore
    
    ld a,l
    or a
    ret nz	// return ( -1, -2 )
    
    in a, (_FDC_STS)
    and #FDC_STS_NOT_RDY | FDC_STS_TRK00
    xor #FDC_STS_NOT_RDY
    ret z	// return ( 0 )
    
    ld l,#0xfd	// return ( -3 )
    ret
    
    __endasm;

/*
    if ( FDC_waitForNotBusy () ) return ( -1 );	// prikaz nelze vlozit
    FDC_CMD = CMD_RESTORE | NO_VERIFY_TRK | HSPEED_30MS;
    if ( FDC_waitForNotBusy () ) return ( -2 ); // prikaz ukoncen s chybou
    if ( FDC_STS_NOT_RDY ^ ( FDC_STS & ( FDC_STS_NOT_RDY | FDC_STS_TRK00 ) ) ) return ( -3 ); // nepresunuli jsme se na TRK00
    return ( 0 );
*/
}





/*
 * FDC_doStepIn - krok nahoru
 *
 * Vraci:
 *	 0: vse OK
 *	-1: prikaz nelze vlozit
 *	-2: prikaz ukoncen s chybou
*/
char FDC_doStepIn ( void ) __naked {

    __asm
    
    ld c,#CMD_STEP_IN | DEF_TRACK | NO_VERIFY_TRK | HSPEED_30MS
    jr __FDC_StepInOutRestore
    
    __endasm;
    
/*
    if ( FDC_waitForNotBusy () ) return ( -1 ); // prikaz nelze spustit
    FDC_CMD = CMD_STEP_IN | DEF_TRACK | NO_VERIFY_TRK | HSPEED_30MS;
    if ( FDC_waitForNotBusy () ) return ( -2 ); // prikaz ukoncen s chybou
    return ( 0 );
*/
}





/*
 * FDC_doStepOut - krok dolu
 *
 * Vraci:
 *	 0: vse OK
 *	-1: prikaz nelze vlozit
 *	-2: prikaz ukoncen s chybou
*/
char FDC_doStepOut ( void ) __naked {

    __asm
    
    ld c,#CMD_STEP_OUT | DEF_TRACK | NO_VERIFY_TRK | HSPEED_30MS
    jr __FDC_StepInOutRestore
    
    __endasm;

/*
    if ( FDC_waitForNotBusy () ) return ( -1 ); // prikaz nelze spustit
    FDC_CMD = CMD_STEP_OUT | DEF_TRACK | NO_VERIFY_TRK | HSPEED_30MS;
    if ( FDC_waitForNotBusy () ) return ( -2 ); // prikaz ukoncen s chybou
    return ( 0 );
*/
}






/*
 * FDC_doSeek - nastaveni stopy
 *
 * Vraci:
 *	 0: vse OK
 *	-1: prikaz nelze vlozit
 *	-2: prikaz ukoncen s chybou
*/
char FDC_doSeek ( unsigned char track ) __naked {
    (void) track;
    __asm
    
    pop de
    pop bc
    push bc
    push de

    call _FDC_waitForNotBusy
    ret nc	// return ( -1 )

    ld a,c
    cpl
    out (_FDC_DATA), a
    
    ld a,#CMD_SEEK | NO_VERIFY_TRK | HSPEED_30MS
    out (_FDC_CMD), a
    
    call _FDC_waitForNotBusy
    ld a,l
    or a
    ret z	// return ( 0 )

    dec l	// return ( -2 )
    ret

    __endasm;
    
/*
    if ( FDC_waitForNotBusy () ) return ( -1 ); // prikaz nelze spustit
    FDC_DATA = ~track;
    FDC_CMD = CMD_SEEK | NO_VERIFY_TRK | HSPEED_30MS;
    if ( FDC_waitForNotBusy () ) return ( -2 ); // prikaz ukoncen s chybou
    return ( 0 );
*/
}





/*
 * FDC_ReadWriteDATA -  
 *		    Vykonani prikazu, ktery cte/zapisuje pres FDC_DATA
 * 		    READ_SECTOR, WRITE_SECTOR a READ_ADDR
 *
 * 		    Pokud nesouhlasi pocet prectenych dat, tak se pokusime
 *		    operaci 5x zopakovat a potom zpanikarime.
 *
 * Vstup:
 *	command 	- FDC prikaz
 *	data_size	- pocet bajtu, ktere se budou prenaset
 *	p		- DMA adresa
 *	drq_addr	- adresa rutiny, ktera bude vykonana pri DRQ
 *
 * Vraci:
 *	0 - pokud je vse OK, nebo cislo chyby
 * 
*/
char FDC_ReadWriteData ( unsigned char command, unsigned int data_size, void *p, void *drq_addr ) {
    char retval = -4;	// chybovy kod pokud se operace nezdarila 5x po sobe
    unsigned char attempt = 5;
    unsigned int read_size;


#ifdef FDC_SUPPORT_HDD
    // TODO: tady musime zakazat interrupt PIO-Z80
    IM1;
    FDC_EINT = EINT_ON;
#endif
    DI;
    
    while ( attempt-- ) {

        if ( FDC_waitForNotBusy () ) {
            retval = -1;	// prikaz nelze spustit
            break;
        };
        
        FDC_CMD = command;

        if ( FDC_waitForFinishCommand () ) {
            retval = -2;	// prikaz neprobehnul spravne
            break;
        };
    
        read_size = FDC_waitForData ( p, drq_addr );
    
    
        if ( FDC_STS != 0xff ) {
            retval = -3;	// vsechny status bity ted musi byt nulove!
            break;
        };
    
        if ( data_size == read_size ) {
            retval = 0;
            break;
        };
    };


#ifdef FDC_SUPPORT_HDD
    // TODO: tady musime povolit interrupt PIO-Z80 a nastavit IM2
    FDC_EINT = EINT_OFF;
#endif
    EI;
    return ( retval );
}






/*
 * FDC_doReadSector - precteni jednoho sektoru
 *
 * Vstup:
 *	sector_size	- pocet bajtu v sektoru
 *	p		- DMA pro ulozeni prectenych dat
 *	inv_data	- 1: data budou invertovany, 0: data se prenesou v surove podobe
 *
 * Vraci:
 *	 0: vse OK
 *	-1: prikaz nelze vlozit
 *	-2: prikaz ukoncen s chybou
 *	-3: vykonani prikazu zastaveno s nejakym neocekavanym priznakem
 *	-4: 5x se nepodarilo prenest vsechny data
*/
char FDC_doReadSector ( unsigned int sector_size,  void *p, unsigned char inv_data ) {
    unsigned char command;
    void *drq_addr;
    
    command = gSIDE | ( CMD_READ_SEC | SINGLE_SECTOR | NO_DELAY15MS | VERIFY_SIDE | DATA_SIGN_FB );
//    command = CMD_READ_SEC | SINGLE_SECTOR | NO_DELAY15MS | NO_VERIFY_SIDE | DATA_SIGN_FB;

    if ( inv_data ) {
        drq_addr = _FDC_readByteCpl;
    } else {
        drq_addr = _FDC_readByte;
    };
    return ( FDC_ReadWriteData ( command, sector_size, p, drq_addr ) );
}




/*
 * FDC_doWriteSector - zapis jednoho sektoru
 *
 * Vstup:
 *	sector_size	- pocet bajtu v sektoru
 *	p		- DMA ze ktere se vyzvednou data pro zapis
 *	inv_data	- 1: data budou invertovany, 0: data se prenesou v surove podobe
 *
 * Vraci:
 *	 0: vse OK
 *	-1: prikaz nelze vlozit
 *	-2: prikaz ukoncen s chybou
 *	-3: vykonani prikazu zastaveno s nejakym neocekavanym priznakem
 *	-4: 5x se nepodarilo prenest vsechny data
*/
char FDC_doWriteSector ( unsigned int sector_size,  void *p, unsigned char inv_data ) {
    unsigned char command;
    void *drq_addr;
    
    command = gSIDE | ( CMD_WRITE_SEC | SINGLE_SECTOR | NO_DELAY15MS | VERIFY_SIDE | DATA_SIGN_FB );
//    command = CMD_READ_SEC | SINGLE_SECTOR | NO_DELAY15MS | NO_VERIFY_SIDE | DATA_SIGN_FB;

    if ( inv_data ) {
        drq_addr = _FDC_writeByteCpl;
    } else {
        drq_addr = _FDC_writeByte;
    };
    return ( FDC_ReadWriteData ( command, sector_size, p, drq_addr ) );
}






/*
 * FDC_doReadAddr - precist 6 bajtu adresy sektoru
 *
 * Vraci:
 *	 0: vse OK
 *	-1: prikaz nelze vlozit
 *	-2: prikaz ukoncen s chybou
 *	-3: vykonani prikazu zastaveno s nejakym neocekavanym priznakem
 *	-4: 5x se nepodarilo prenest vsechny data
*/
char FDC_doReadAddr ( DISCADDR_t *p ) {
    unsigned char command;
    
    command = CMD_READ_ADDR | NO_DELAY15MS;
    return ( FDC_ReadWriteData ( command, 6, p, _FDC_readByteCpl ) );
}

