
#ifndef _driver_FDC
#define _driver_FDC

//#define FDC_SUPPORT_HDD

/* struktura dat, ktere jsou precteny FDC prikazem READ_ADDR */
typedef struct {
    unsigned char track;
    unsigned char sector;
    unsigned char side;
    unsigned char sector_size;
    unsigned int crc;
} DISCADDR_t;


#define FDC_NOT_INVERT_DATA	0
#define FDC_INVERT_DATA		1

#define FDC_DENSITY_DD		0
#define FDC_DENSITY_HD		1

#define FDC_DRIVE0		0x00
#define FDC_DRIVE1		0x01
#define FDC_DRIVE2		0x02
#define FDC_DRIVE3		0x03

/* invertovane hodnoty FDC_STS */
#define FDC_STS_NOT_RDY		( 1 << 7 )
#define FDC_STS_WR_PRT		( 1 << 6 ) /* jen po I., CMD_WRITE_* a CMD_WRITE_TRACK */
#define FDC_STS_HEAD_LOAD	( 1 << 5 ) /* jen po I., sdeluje, ze je hlava prilozena k mediu */
#define FDC_STS_REC_TYPE	( 1 << 5 ) /* jen po CMD_READ*, ??? */
#define FDC_STS_SEEK_ERR	( 1 << 4 ) /* jen po I. */
#define FDC_STS_RNF		( 1 << 4 ) /* CMD_READ_ADDR, CMD_READ* a CMD_WRITE_* */
#define FDC_STS_CRC_ERR		( 1 << 3 )
#define FDC_STS_TRK00		( 1 << 2 ) /* jen po I., hlava je nastavena nad nultou stopou */
#define FDC_STS_DATA_LOST	( 1 << 2 )
#define FDC_STS_INDEX		( 1 << 1 ) /* jen po I. */
#define FDC_STS_DRQ		( 1 << 1 )
#define FDC_STS_BUSY		( 1 << 0 )

/* prace s FDC registry */
extern void FDC_setTrack ( unsigned char track );
extern unsigned char FDC_getTrack ( void );
extern void FDC_setSector ( unsigned char sector );
extern unsigned char FDC_getSector ( void );
extern void FDC_setSide ( unsigned char side );
extern unsigned char FDC_getStatus ( void );
#ifdef FDC_SUPPORT_HDD
extern void FDC_setDensity ( unsigned char density );
#endif

/* prikazy mechaniky */
extern void FDC_doMotorOn ( void );
extern void FDC_doMotorOff ( void );
extern void FDC_doSelectDrive ( char drive_id );


/* krokovaci prikazy */
extern char FDC_doRestore ( void );
extern char FDC_doStepIn ( void );
extern char FDC_doStepOut ( void );
extern char FDC_doSeek ( unsigned char track );

/* cteni/zapis sektoru */
char FDC_doReadSector ( unsigned int sector_size,  void *p, unsigned char inv_data );
char FDC_doWriteSector ( unsigned int sector_size,  void *p, unsigned char inv_data );


/* specialni prikazy */
extern char FDC_doReadAddr ( DISCADDR_t *p );

#endif
