
#include <stdio.h>

#include "video.h"
#include "monitor.h"

#include "test_FDC.h"
#include "driver_FDC.h"

__sfr __at 		0xce 	rDMD;

// porty k mapovani standardni pameti
__sfr __at 		0xe0 	rMAPE0;
__sfr __at 		0xe1 	rMAPE1;
__sfr __at 		0xe2 	rMAPE2;
__sfr __at 		0xe3 	rMAPE3;
__sfr __at 		0xe4 	rMAPE4;
__sfr __at 		0xe5 	rMAPE5;
__sfr __at 		0xe6 	rMAPE6;



void prg_error ( unsigned char err_code ) {
    xprintf ( " ERR = %d", err_code );
    while ( 1 ) {
        rBCOLOR = err_code;
        rBCOLOR = 0x00;
    };
}


void print_result ( char *s ) {
    cur.pos_x = 15;
    MZ800_INVPRNon ();
    prnc ( '[' );
    msg ( s );
    prnc ( ']' ); 
    MZ800_INVPRNoff ();
}

void main ( void ) {

    char err = 0;

    unsigned char drive_id = FDC_DRIVE0;

    /* testovaci hodnoty pro Seek a readSector */
    unsigned char track = 10;
    unsigned char sector = 2;
    unsigned char side = 1;

    unsigned int sector_size;
    void *dma_addr = (void *) 0x6000;
    DISCADDR_t sector_addr;


    xputc = prnc;
    xputs = msg;

    rBCOLOR = BLACK;
    rPAL = PAL0 | BLACK;
    rPAL = PAL1 | LWHITE;
    rPAL = PAL2 | WHITE;
    rPAL = PAL3 | BLACK;
    
    rWF = PLANE_I | PLANE_III;
    video_cls ();
    rWF = PLANE_I;
    MZ800_INVPRNoff ();
    cur.pos_x = 0;
    cur.pos_y = 0;



    /* mame pripojen radic? */
    xprintf ( "FDC test:" );

    if ( 0 == testFDC_isPresent () ) {
        print_result ( "OK" );
#ifdef FDC_SUPPORT_HDD
        xputs ( " - HD driver" );
#else
        xputs ( " - normal driver" );
#endif        
    } else {
        print_result ( "NOT FOUND" );
        // radic nenalezen, nema smysl dal pokracovat
        while ( 1 ) {};
    };




    /* roztocit motor a vybrat mechaniku */
    cur.pos_x = 0;
    cur.pos_y++; 
    xprintf ( "Select FD%x:", drive_id&3 );

    FDC_doMotorOn ();
    FDC_doSelectDrive ( drive_id );

    print_result ( "SELECTED" );




    /* provedeme restorte (TRK00) */
    cur.pos_x = 0;
    cur.pos_y++;
    xprintf ( "Set TRK00:" );

    err = FDC_doRestore ();

    if ( err ) prg_error ( err );
    print_result ( "OK" );



    /* vyzkousime step in (nahoru) / out (dolu) */
    cur.pos_x = 0;
    cur.pos_y++;
    xprintf ( "Step IN, OUT:" );
    
    err = FDC_doStepIn ();

    if ( err ) prg_error ( err );
    print_result ( "OK" );
    msg ( ", " );

    err = FDC_doStepOut ();

    if ( err ) prg_error ( err );
    MZ800_INVPRNon ();
    msg ( "[OK]" );
    MZ800_INVPRNoff ();




    /* seek na pozadovanou stopu */
    cur.pos_x = 0;
    cur.pos_y++;

    xprintf ( "Seek %d:", track );

    err = FDC_doSeek ( track );

    if ( err ) prg_error ( err );
/*
    // tohle asi neni uplne spravne testovano
    if ( track != FDC_getTrack () ) {
        print_result ( "FAILED" );
        while ( 1 ) {};
    };
*/
    print_result ( "OK" );




    /* precteni FDC adresy */
    cur.pos_x = 0;
    cur.pos_y++;
    xprintf ( "Read addr:" );

    FDC_setSide ( side );
    FDC_setSector ( sector );
    err = FDC_doReadAddr ( &sector_addr );

    if ( err ) prg_error ( err );
    print_result ( "OK" );
    xprintf (" - trk=0x%02x, sec=0x%02x, side=0x%02x, size=0x%02x, crc=0x%04x", sector_addr.track, sector_addr.sector, sector_addr.side&1, sector_addr.sector_size, sector_addr.crc);




    /* precteni sektoru */
    cur.pos_x = 0;
    cur.pos_y += 2;

    /* velikost sektoru si ted nacteme z FDC, jinak by mela byt nastavena nekde v konfiguracni tabulce */
    if ( sector_addr.sector_size == 0 ) {
        sector_size = 128;
    } else if ( sector_addr.sector_size == 1 ) {
        sector_size = 256;
    } else {
        sector_size = 512;
    };
    xprintf ( "DMA = 0x%04x, sector_size = %d bytes...", (unsigned int) dma_addr, sector_size );
    
    cur.pos_x = 0;
    cur.pos_y++;    
    xprintf ( "Read sector:" );


#ifdef FDC_SUPPORT_HDD
    FDC_setDensity ( FDC_DENSITY_DD );
#endif
    err = FDC_doReadSector ( sector_size, dma_addr, FDC_INVERT_DATA );

    if ( err ) prg_error ( err );
    print_result ( "OK" );





    FDC_doMotorOff ();
    
    
    while ( 1 ) {};
}
